!! A utility class for 1D TFWI SEPlib IO
!!
!! @author Biondo Biondi
!! @version 2016.07.15
module Util1Daxis_mod

  use sep
  use axis_mod
  use mem_mod

  implicit none
  logical, private, parameter    :: T = .true., F = .false.

contains

  function get_backslow_axis(tag) result(saxis)
    type(axis)             :: saxis
    integer                :: ns
    real                   :: os
    real                   :: ds
    character(*), optional :: tag

    if(present(tag)) then
      call from_aux(tag, "n1", ns, 1)
      call from_aux(tag, "o1", os, 0.)
      call from_aux(tag, "d1", ds, 0.)
    else
      call from_param("ns", ns, 1)
      call from_param("os", os, 0.)
      call from_param("ds", ds, 1.)
    endif

    saxis = axis(ns, os, ds)
  end function

  function get_slow_axis(tag) result(lagaxis)
    type(axis)                :: lagaxis
    character(*), optional    :: tag           !! name of history file
    integer                   :: lag_n         !! lag axis count
    real                      :: lag_o         !! lag axis origin
    real                      :: lag_d         !! lag axis sampling
    integer                   :: slow_phys_lag !! physical slowness lag count
    integer                   :: zero_lag      !! location of zero lag

    if(present(tag)) then
      call from_aux(tag,"n1", lag_n)
      call from_aux(tag,"o1", lag_o)
      call from_aux(tag,"d1", lag_d)
    else
      call from_param("ntau", lag_n)
      call from_param("otau", lag_o)
      call from_param("dtau", lag_d)
    endif

    lagaxis = axis(lag_n,lag_o,lag_d)

    if (lag_n .ne. 1) then
      slow_phys_lag=lag_n
      zero_lag=nint(-lag_o/lag_d+1)
      if(zero_lag < 1) call seperr("zero lag is too small")
      if(zero_lag > lag_n-1) call seperr("zero lag is too large")
    else
      slow_phys_lag=1
    end if

  end function

  !! Gets the axes of the input wavelet either from a history or par file
  subroutine get_wavelet_axis(waxis, wavelet_tdelay, tag)
    type(axis)            :: waxis               !! axis object
    integer               :: wavelet_n           !! time axis count
    real                  :: wavelet_o,wavelet_d !! time axis sampling
    real                  :: wavelet_tdelay      !! wavelet time delay
    character(*),optional :: tag                 !! wavelet tag

    if(present(tag)) then
      call from_aux(tag, "n1", wavelet_n)
      call from_aux(tag, "o1", wavelet_o)
      call from_aux(tag, "d1", wavelet_d)
      call from_aux(tag, "tdelay", wavelet_tdelay)
    else
      call from_param("ntw", wavelet_n)
      call from_param("otw", wavelet_o,0.)
      call from_param("dtw", wavelet_d)
      call from_param("dly", wavelet_tdelay)
    endif

    waxis = axis(wavelet_n,wavelet_o,wavelet_d)

  end subroutine

  !! Gets the axes of the input wavelet either from a history or par file
  function get_data_axis(tag) result(daxis)
    type(axis)            :: daxis !! axis object
    integer               :: t_n   !! time axis count
    real                  :: t_o   !! time axis origin
    real                  :: t_d   !! time axis sampling
    character(*),optional :: tag   !! data tag

    if(present(tag)) then
      call from_aux(tag, "n1", t_n)
      call from_aux(tag, "o1", t_o)
      call from_aux(tag, "d1", t_d)
    else
      call from_param("nt", t_n)
      call from_param("ot", t_o)
      call from_param("dt", t_d)
    endif

    daxis = axis(t_n,t_o,t_d)

  end function

  subroutine put_wavelet_axis(waxis, tag, saxis)
    type(axis)           :: waxis
    type(axis), optional :: saxis
    character(*)         :: tag

    call to_history("n1", waxis%n, tag)
    call to_history("o1", waxis%o, tag)
    call to_history("d1", waxis%d, tag)
    call to_history("label1", "Time (s)", tag)
    if(present(saxis)) then
      call to_history("n2", saxis%n, tag)
      call to_history("o2", saxis%o, tag)
      call to_history("d2", saxis%d, tag)
      call to_history("label2", "Slowness (s/km)", tag)
    endif
  end subroutine

  subroutine put_data_axis(daxis, tag, saxis)
    type(axis)           :: daxis
    type(axis), optional :: saxis
    character(*)         :: tag

    call to_history("n1", daxis%n, tag)
    call to_history("o1", daxis%o, tag)
    call to_history("d1", daxis%d, tag)
    call to_history("label1", "Time (s)", tag)
    if(present(saxis)) then
      call to_history("n2", saxis%n, tag)
      call to_history("o2", saxis%o, tag)
      call to_history("d2", saxis%d, tag)
      if(saxis%d == 0.0) call to_history("d2", 0.01, tag)
      call to_history("label2", "Slowness (s/km)", tag)
    endif
  end subroutine

  !! Puts the slowness axes to the output slowness history or parameter file
  subroutine put_slow_axis(caxis, tag, saxis)
    character(*)         :: tag    !! name of history file
    type(axis)           :: caxis
    type(axis), optional :: saxis

    if(present(saxis)) then
      call to_history("n1", caxis%n, tag)
      call to_history("o1", caxis%o, tag)
      call to_history("d1", caxis%d, tag)
      call to_history("label1", "lag (s)", tag)
      call to_history("n2", saxis%n, tag)
      call to_history("o2", saxis%o, tag)
      call to_history("d2", saxis%d, tag)
      call to_history("label2", "Slowness (s/km)", tag)
    else
      call to_history("n1", caxis%n, tag)
      call to_history("o1", caxis%o, tag)
      call to_history("d1", caxis%d, tag)
      call to_history("label1", "lag (s)", tag)
    endif
  end subroutine

  function calc_zlag(caxis) result(zlag)
    type(axis)     :: caxis
    integer        :: zlag, lag_n
    real           :: lag_o, lag_d

    lag_n = caxis%n; lag_o = caxis%o
    lag_d = caxis%d
 
    zlag = nint(-lag_o/lag_d+1)
    if(zlag < 1) call seperr("zero lag is too small")
    if(zlag > lag_n - 1) call seperr("zero lag is too large")

  end function

end module
