module nlsolver_mod
! A driver for the nonlinear solver library
! Last edited by Ali Almomin 08/24/2015

  use sep
  use solver_obj_mod
  use solver_nlcg_mod
  use solver_lcg_mod
  use solver_lbfgs_mod
  use stepper_obj_mod
  use stepper_backtrack_mod
  use stepper_brent_mod
  use stepper_parab_mod
  use stepper_sample_mod
  use stepper_linear_mod
  use terminate_obj_mod
  use terminate_simple_mod
  use prob_obj_mod

  implicit none
!-------------------- module parameters --------------------
  integer, parameter                :: NLCG = 11
  integer, parameter                :: BFGS = 12
  integer, parameter                :: LCG = 13
  integer, parameter                :: BT = 21
  integer, parameter                :: BRENT = 22
  integer, parameter                :: PARAB = 23
  integer, parameter                :: SAMPLE = 24
  integer, parameter                :: LINEAR = 25

  contains
!-------------------- driver function --------------------
  integer function nlsolver(modl, prob, nsolver, nstepper, niter, solver, stepper, term)
    optional ::                         nsolver, nstepper, niter, solver, stepper, term
!-------------------- interface variables --------------------
    real, dimension(:)              :: modl          ! model (input/output)
    class(prob_obj)                 :: prob          ! problem object (input)
    integer                         :: nsolver       ! solver number (input)
    integer                         :: nstepper      ! stepper number (input)
    integer                         :: niter         ! number of iterations (input)
    class(solver_obj), target       :: solver        ! solver object (input)
    class(stepper_obj), target      :: stepper       ! stepper object (input)
    class(terminate_obj), target    :: term          ! terminator object (input)
!-------------------- function variables --------------------
    integer                         :: stat          ! status flag
    integer                         :: nsolverl      ! local solver number
    integer                         :: nstepperl     ! local stepper number
    integer                         :: niterl        ! local number of iterations
    class(solver_obj), pointer      :: solverl       ! local solver object
    class(stepper_obj), pointer     :: stepperl      ! local stepper object
    class(terminate_obj), pointer   :: terml         ! local terminator object
    type(solver_nlcg), target       :: solver1       ! NLCG solver
    type(solver_lbfgs), target      :: solver2       ! LBFGS solver
    type(solver_lcg), target        :: solver3       ! LCG solver
    type(stepper_backtrack), target :: stepper1      ! Backtrack stepper
    type(stepper_brent), target     :: stepper2      ! Brent stepper
    type(stepper_parab), target     :: stepper3      ! Parabola stepper
    type(stepper_sample), target    :: stepper4      ! Sampling stepper
    type(stepper_linear), target    :: stepper5      ! Linear stepper
    type(terminate_simple), target  :: term1         ! terminate_obj object
!-------------------- set solver --------------------
    if(present(solver)) then
      solverl => solver
    else
      nsolverl = NLCG
      if(present(nsolver)) nsolverl = nsolver
      if(nsolverl .eq. NLCG) then
        solver1 = solver_nlcg()
        solverl => solver1
      else if(nsolverl .eq. BFGS) then
        solver2 = solver_lbfgs()
        solverl => solver2
      else if(nsolverl .eq. LCG) then
        solver3 = solver_lcg()
        solverl => solver3
      end if
    end if
!-------------------- set stepper --------------------
    if(present(stepper)) then
      stepperl => stepper
    else
      nstepperl = BT
      if(present(nstepper)) nstepperl = nstepper
      if(nstepperl .eq. BT) then
        stepper1 = stepper_backtrack()
        stepperl => stepper1
      else if(nstepperl .eq. BRENT) then
        stepper2 = stepper_brent()
        stepperl => stepper2
      else if(nstepperl .eq. PARAB) then
        stepper3 = stepper_parab()
        stepperl => stepper3
      else if(nstepperl .eq. SAMPLE) then
        stepper4 = stepper_sample()
        stepperl => stepper4
      else if(nstepperl .eq. LINEAR) then
        stepper5 = stepper_linear()
        stepperl => stepper5
      end if
    end if
!-------------------- set terminator --------------------
    if(present(term)) then
      terml => term
    else
      niterl = 10
      if(present(niter)) niterl = niter
      term1 = terminate_simple(niter=niterl)
      terml => term1
    end if
!-------------------- run solver --------------------
    stat = solverl%run(modl, prob, stepperl, terml)
    nlsolver = 0
  end function

end module
