module prob_nl_l2_diff_mod
! Nonlinear L2 difference objective form
! Last edited by Ali Almomin 08/24/2015
! 0 = Wd (L(S Wm m) - d)

  use sep
  use prob_obj_mod

  implicit none

  private
  public :: prob_nl_l2_diff

  logical, parameter                           :: T = .true., F = .false.

  type, extends(prob_obj)                       :: prob_nl_l2_diff
    private
    real, dimension(:), pointer        :: data
    procedure(oper), nopass, pointer   :: nlop
    procedure(update), nopass, pointer :: uop
    procedure(oper), nopass, pointer   :: lop
  contains
    private
    procedure                                  :: objf
    procedure                                  :: resf
    procedure                                  :: gradf
    procedure                                  :: dresf
  end type

  interface prob_nl_l2_diff
    module procedure const
  end interface

  interface
    integer function oper(adj, add, modl, data)
      real,dimension(:) :: modl,data
      logical           :: adj,add
    end function
    integer function update(modl)
      real :: modl
    end function
  end interface

  contains

  type(prob_nl_l2_diff) function const(nlop, uop, lop, nmodl, ndata, data)
    procedure(oper)              :: nlop, lop
    procedure(update)            :: uop
    integer                      :: nmodl          ! size of modl (input)
    integer                      :: ndata          ! size of data (input)
    integer                      :: stat           ! 
    real, dimension(:), target   :: data           ! data (input)
    const%nlop => nlop
    const%uop => uop
    const%lop => lop
    const%nmodl = nmodl
    const%ndata = ndata
    const%data => data
    stat = const%alloc()
  end function

  integer function objf(this, res, obj)   ! function for calculating objective function
    class(prob_nl_l2_diff)   :: this
    real, dimension(:)       :: res  ! residual (input)
    double precision         :: obj  ! objective function value (output)
    obj = 0.5*(sum(dprod(res,res)))
    objf = 0
  end function

  integer function resf(this, modl, res)  ! function for calculating the residual
    class(prob_nl_l2_diff) :: this
    real, dimension(:)     :: modl  ! model (input)
    real, dimension(:)     :: res   ! residual (output)
    integer                :: stat  ! status flag
    stat = this%nlop(F, F, modl, res)
    res = res - this%data
    resf = 0
  end function

  integer function gradf(this, modl, res, grad) ! function for calculating gradient
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (input)
    real, dimension(:) :: grad            ! gradient (output)
    integer            :: stat            ! status flag
    stat = this%uop(modl(1))
    stat = this%lop(T, F, grad, res)
    gradf = 0
  end function

  integer function dresf(this, modl, dmodl, dres) ! function for calculating gradient in data space
    class(prob_nl_l2_diff)        :: this
    real, dimension(:) :: modl           ! model (input)
    real, dimension(:) :: dmodl          ! gradient (input)
    real, dimension(:) :: dres           ! gradient in data space (output)
    integer            :: stat           ! status flag
    stat = this%uop(modl(1))
    stat = this%lop(F, F, dmodl, dres)
    dresf = 0
  end function

end module
