module prob_nl_l2_diff_ext_mod
! Nonlinear L2 difference objective form
! for 1D simple EFWI
! Last edited by Ali Almomin 08/24/2015
! 0 = Wd (L(S Wm m) - d)

  use sep
  use prob_obj_mod

  implicit none

  private
  public :: prob_nl_l2_diff_ext

  logical, parameter                           :: T = .true., F = .false.

  type, extends(prob_obj)                       :: prob_nl_l2_diff_ext
    private
    real, dimension(:), pointer        :: data
    real, dimension(:), allocatable    :: tmp_modl
    real, dimension(:), allocatable    :: tmp_data
    procedure(oper),      nopass, pointer :: nlop
    procedure(oper),      nopass, pointer :: lop
    procedure(oper),      nopass, pointer :: lopext
    procedure(oper),      nopass, pointer :: nlopext
    procedure(update),    nopass, pointer :: uopl
    procedure(update),    nopass, pointer :: uoplextbk
    procedure(extupdate), nopass, pointer :: uoplext
    procedure(update),    nopass, pointer :: uopnlext
  contains
    !private
    procedure                                  :: objf
    procedure                                  :: resf
    procedure                                  :: gradf
    procedure                                  :: dresf
  end type

  interface prob_nl_l2_diff_ext
    module procedure const
  end interface

  interface
    integer function oper(adj, add, modl, data)
      real,dimension(:) :: modl,data
      logical           :: adj,add
    end function
    integer function update(modl)
      real :: modl
    end function
    integer function extupdate(modl)
      real,dimension(:) :: modl
    end function
  end interface

  contains

  type(prob_nl_l2_diff_ext) function const(nlop, nlopext, uopl, uoplext, uoplextbk, uopnlext, &
                                           lop, lopext, nmodl, ndata, data)
    procedure(oper)              :: nlop, nlopext, lop, lopext
    procedure(update)            :: uopl, uoplextbk, uopnlext
    procedure(extupdate)         :: uoplext
    integer                      :: nmodl          ! size of modl (input)
    integer                      :: ndata          ! size of data (input)
    integer                      :: stat           ! 
    real, dimension(:), target   :: data           ! data (input)
    const%nlop      => nlop
    const%nlopext   => nlopext
    const%uopl      => uopl
    const%uoplext   => uoplext
    const%uoplextbk => uoplextbk
    const%uopnlext  => uopnlext
    const%lop       => lop
    const %lopext   => lopext
    const%data      => data
    const%nmodl     = nmodl
    const%ndata     = ndata
    allocate(const%tmp_data(const%ndata))
    allocate(const%tmp_modl(const%nmodl))
    stat = const%alloc()
  end function

  integer function objf(this, res, obj)   ! function for calculating objective function
    class(prob_nl_l2_diff_ext)   :: this
    real, dimension(:)       :: res  ! residual (input)
    double precision         :: obj  ! objective function value (output)
    obj = 0.5*(sum(dprod(res,res)))
    objf = 0
  end function

  integer function resf(this, modl, res)  ! function for calculating the residual
    class(prob_nl_l2_diff_ext) :: this
    real, dimension(:)     :: modl  ! model (input)
    real, dimension(:)     :: res   ! residual (output)
    integer                :: stat  ! status flag
    stat = this%uopnlext(modl(this%nmodl))
    stat = this%nlop    (F, F, modl, this%tmp_data)
    stat = this%nlopext (F, F, modl, res)
    res = res + this%tmp_data
    res = res - this%data
    resf = 0
  end function

  integer function gradf(this, modl, res, grad) ! function for calculating gradient
    class(prob_nl_l2_diff_ext)        :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (input)
    real, dimension(:) :: grad            ! gradient (output)
    integer            :: stat            ! status flag
    integer            :: test
    ! Update operators
    stat = this%uopl      (modl(this%nmodl      ))
    stat = this%uoplextbk (modl(this%nmodl      ))
    stat = this%uoplext   (modl(1:this%nmodl - 1))
    stat = this%uopnlext  (modl(this%nmodl      ))
    ! Compute gradient in s
    stat = this%lop   (T, F, grad, res)
    stat = this%lopext(T, F, this%tmp_modl, res)
    grad(this%nmodl) = grad(this%nmodl) + this%tmp_modl(this%nmodl)
    ! Compute gradient in c
    stat = this%nlopext(T, F, this%tmp_modl, res)
    grad(1:this%nmodl-1) = this%tmp_modl(1:this%nmodl-1)
    gradf = 0
  end function

  integer function dresf(this, modl, dmodl, dres) ! function for calculating gradient in data space
    class(prob_nl_l2_diff_ext)        :: this
    real, dimension(:) :: modl           ! model (input)
    real, dimension(:) :: dmodl          ! gradient (input)
    real, dimension(:) :: dres           ! gradient in data space (output)
    integer            :: stat           ! status flag
    ! Update operators
    stat = this%uopl      (modl(this%nmodl      ))
    stat = this%uoplextbk (modl(this%nmodl      ))
    stat = this%uoplext   (modl(1:this%nmodl - 1))
    stat = this%uopnlext  (modl(this%nmodl      ))
    ! Conjugate gradient in s
    stat = this%lop(F, F, dmodl, dres)
    stat = this%lopext(F, F, modl, this%tmp_data)
    dres = dres + this%tmp_data
    ! Conjugate gradient in c
    stat = this%nlopext(F, F, dmodl, this%tmp_data)
    dres = dres + this%tmp_data
    dresf = 0
  end function

end module
