module prob_obj_mod
! Problem abstract object
! Last edited by Ali Almomin 08/24/2015

  use sep

  implicit none

  private
  public :: prob_obj

  type, abstract :: prob_obj
    private
    integer, public                            :: nmodl
    integer, public                            :: ndata
    integer                                    :: fevals = 0
    integer,public                             :: counter = 0
    logical                                    :: obj_updated = .false.
    logical                                    :: res_updated = .false.
    logical                                    :: grad_updated = .false.
    logical                                    :: dres_updated = .false.
    double precision                           :: obj = 0.
    real, dimension(:), allocatable            :: modl
    real, dimension(:), allocatable,public     :: res
    real, dimension(:), allocatable            :: grad
    real, dimension(:), allocatable            :: dmodl
    real, dimension(:), allocatable,public     :: dres
    character(256), public                     :: rnorm = ''         ! tag for outputting residual norm
    character(256), public                     :: mmov = ''          ! tag for outputting models movie
    character(256), public                     :: gmov = ''          ! tag for outputting gradient movie
    character(256), public                     :: rmov = ''          ! tag for outputting residuals movie
  contains
    procedure                                           :: objf
    procedure                                           :: resf
    procedure                                           :: gradf
    procedure                                           :: dresf
    procedure                                           :: datresout
    procedure                                           :: alloc
    procedure                                           :: set_modl
    procedure                                           :: get_obj
    procedure                                           :: get_res
    procedure                                           :: get_grad
    procedure                                           :: get_dres
    procedure                                           :: get_fevals
    procedure,public                                    :: get_counter
    procedure                                           :: get_rnorm
    procedure                                           :: get_gnorm
    procedure                                           :: output
  end type

  contains

  integer function alloc(this)
    class(prob_obj)              :: this
    if(this%nmodl < 1 .or. this%ndata < 1) call erexit("problem: nmodl or ndata is not correct")
    allocate(this%modl(this%nmodl))
    allocate(this%res(this%ndata))
    allocate(this%grad(this%nmodl))
    allocate(this%dres(this%ndata))
    allocate(this%dmodl(this%nmodl))
    alloc = 0
  end function

  integer function set_modl(this, modl)
    class(prob_obj)              :: this
    real, dimension(:)          :: modl
    if(any(this%modl .ne. modl)) then
      this%modl = modl
      this%obj_updated = .false.
      this%res_updated = .false.
      this%grad_updated = .false.
      this%dres_updated = .false.
    end if
    set_modl = 0
  end function

  integer function get_obj(this, modl, obj)
    class(prob_obj)              :: this
    real, dimension(:)          :: modl
    double precision            :: obj
    integer                     :: stat
    if(size(modl) .ne. this%nmodl) call erexit("problem: size of modl is not correct")
    stat = this%set_modl(modl)
    if(.not. this%obj_updated) then
      stat = this%get_res(this%modl, this%res)
      stat = this%objf(this%res, this%obj)
      this%obj_updated = .true.
    end if
    obj = this%obj
    get_obj = 0
  end function

  integer function get_res(this, modl, res)
    class(prob_obj)              :: this
    real, dimension(:)          :: modl
    real, dimension(:)          :: res
    integer                     :: stat
    if(size(modl) .ne. this%nmodl .or. size(res) .ne. this%ndata) call erexit("problem: size of modl or res is not correct")
    stat = this%set_modl(modl)
    if(.not. this%res_updated) then
      stat = this%resf(this%modl, this%res)
      this%fevals = this%fevals + 1
      this%res_updated = .true.
    end if
    res = this%res
    get_res = 0
  end function

  integer function get_grad(this, modl, grad)
    class(prob_obj)              :: this
    real, dimension(:)          :: modl
    real, dimension(:)          :: grad
    integer                     :: stat
    if(size(modl) .ne. this%nmodl .or. size(grad) .ne. this%nmodl) call erexit("problem: size of modl or res is not correct")
    stat = this%set_modl(modl)
    if(.not. this%grad_updated) then
      stat = this%get_res(this%modl, this%res)
      stat = this%gradf(this%modl, this%res, this%grad)
      this%fevals = this%fevals + 2
      this%grad_updated = .true.
    end if
    grad = this%grad
    get_grad = 0
  end function

  integer function get_dres(this, modl, dmodl, dres)
    class(prob_obj)              :: this
    real, dimension(:)          :: modl
    real, dimension(:)          :: dmodl
    real, dimension(:)          :: dres
    integer                     :: stat
    if(size(modl) .ne. this%nmodl .or. size(dmodl) .ne. this%nmodl .or. size(dres) .ne. this%ndata) call erexit("problem: size of modl or dmodl or dres is not correct")
    stat = this%set_modl(modl)
    if(.not. this%dres_updated .or. any(this%dmodl .ne. dmodl)) then
      this%dmodl = dmodl
      stat = this%dresf(this%modl, this%dmodl, this%dres)
      this%fevals = this%fevals + 2
      this%dres_updated = .true.
    end if
    dres = this%dres
    get_dres = 0
  end function

  integer function get_fevals(this)
    class(prob_obj)              :: this
    get_fevals = this%fevals
  end function

  integer function get_counter(this)
    class(prob_obj)              :: this
    get_counter = this%counter
  end function

  double precision function get_rnorm(this)
    class(prob_obj)              :: this
    integer                     :: stat
    stat = this%get_res(this%modl, this%res)
    get_rnorm = sqrt(sum(dprod(this%res,this%res)))
  end function

  double precision function get_gnorm(this)
    class(prob_obj)              :: this
    integer                     :: stat
    stat = this%get_grad(this%modl, this%grad)
    get_gnorm = sqrt(sum(dprod(this%grad,this%grad)))
  end function

  integer function output(this, modl)
    class(prob_obj)               :: this
    real, dimension(:)           :: modl
    integer                      :: stat
    this%counter = this%counter + 1
    if(this%counter .eq. 1) then
      if(this%rnorm .ne. '') then
        call to_history("label1","iteration",this%rnorm)
      end if
      if(this%mmov .ne. '') then
        call to_history("n1",this%nmodl,this%mmov)
        call to_history("label1","Model index",this%mmov)
        call to_history("label2","iteration",this%mmov)
      end if
      if(this%gmov .ne. '') then
        call to_history("n1",this%nmodl,this%gmov)
        call to_history("label1","Gradient index",this%gmov)
        call to_history("label2","iteration",this%gmov)
      end if
      if(this%rmov .ne. '') then
        call to_history("n1",this%ndata,this%rmov)
        call to_history("label1","Residual index",this%rmov)
        call to_history("label2","iteration",this%rmov)
      end if
    end if
    if(this%rnorm .ne. '') then
      call to_history("n1",this%counter,this%rnorm)
      stat = this%get_obj(modl, this%obj)
      call srite(this%rnorm, real(this%obj), 4)
    end if
    if(this%mmov .ne. '') then
      call to_history("n2",this%counter,this%mmov)
      call srite(this%mmov, modl, 4*size(modl))
    end if
    if(this%gmov .ne. '') then
      call to_history("n2",this%counter,this%gmov)
      stat = this%get_grad(modl, this%grad)
      call srite(this%gmov, this%grad, 4*size(this%grad))
    end if
    if(this%rmov .ne. '') then
      call to_history("n2",this%counter,this%rmov)
      stat = this%get_res(modl, this%res)
      call srite(this%rmov, this%res, 4*size(this%res))
    end if
    ! Write out data residual norm
    stat = this%datresout(modl)
    output = 0
  end function

  integer function objf(this, res, obj)         ! function for calculating objective function
    class(prob_obj)              :: this
    real, dimension(:) :: res             ! residual (input)
    double precision   :: obj             ! objective function value (output)
    call erexit("ERROR: objf was not overridden")
    objf = 0
  end function

  integer function resf(this, modl, res)        ! function for calculating the residual
    class(prob_obj)              :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (output)
    call erexit("ERROR: resf was not overridden")
    resf = 0
  end function

  integer function gradf(this, modl, res, grad) ! function for calculating gradient
    class(prob_obj)              :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: res             ! residual (input)
    real, dimension(:) :: grad            ! gradient (output)
    call erexit("ERROR: gradf was not overridden")
    gradf = 0
  end function

  integer function dresf(this, modl, dmodl, dres) ! function for calculating gradient
    class(prob_obj)              :: this
    real, dimension(:) :: modl            ! model (input)
    real, dimension(:) :: dmodl             ! residual (input)
    real, dimension(:) :: dres            ! gradient (output)
    call erexit("ERROR: dresf was not overridden")
    dresf = 0
  end function

  integer function datresout(this, modl)
    class(prob_obj)                   :: this
    real, dimension(:)                :: modl          ! residual (input)
    double precision                  :: obj1          ! objective function value (output)
    integer                           :: stat          ! status flag
    datresout = 0 
  end function

end module
