!! Linearization of non-linear, non-extended operator
!! <p>
!! Forward takes a perturbation in slowness (scalar) 
!! and scales it by the shifted wavelet derivative
!! <p>
!! Adjoint takes shifted wavelet and computes
!! the perturbation in slowness.
!! <p>
!! L in the notes
!! <p>
!! @author Biondo Biondi
!! @version 2016.07.15
module prop_der11Daxis_mod

  use axis_mod
  use mem_mod
  use adj_mod
  use lint1_stat  
  use igrad1

  implicit none

  logical, private, parameter                :: T = .true., F = .false.

  integer, private                           :: nt                !! data axis one count
  integer, private                           :: nw                 !! wavelet axis count
  integer, private                           :: slow_phys_lag      !! position of physical slowness
  real, private                              :: dist          !! propagation distance (l)
  real, private                              :: ow                 !! wavelet axis origin
  real, private                              :: dw                 !! wavelet axis sampling
  real, private                              :: ot                !! data axis one origin 
  real, private                              :: dt                !! data axis one sampling
  real, private                              :: dly               !! wavelet time delay
  real, private                              :: slow_phys          !! slowness model (scalar)
  real, private                              :: dtdslow            !! negative of propagation distance

  real, dimension(:), allocatable, private   :: wavder       !! first derivative of the wavelet
  real, dimension(:), allocatable, private   :: wavder_shift !! shifted first derivative

  contains

  !! Initializes the operator
  subroutine prop_der11Daxis_init(daxin, waxin, dly_in, dist_in, wavder_in)
    type(axis)         :: daxin, waxin
    real               :: dly_in            !! wavelet delay
    real               :: dist_in                 !! propagation distance
    real, dimension(:) :: wavder_in

    nt = daxin%n; ot = daxin%o; dt = daxin%d
    nw = waxin%n; ow = waxin%o; dw = waxin%d

    dly  = dly_in
    dist = dist_in

    dtdslow = -dist

    call mem_alloc(wavder, nw)
    wavder = wavder_in

    call mem_alloc(wavder_shift, nt)

    if (dw .ne. dt) then
      write(0,*)'dtdata=',dt
      write(0,*)'dtwavelet=',dw
      call seperr("dtwavelet must be equal to dtdata")
    end if
  end subroutine

  !! Updates the slowness
  !! <p>
  !! This function or its corresponding subroutine must be
  !! called each time before applying the operator
  integer function prop_der11Daxis_phys_slow_update(slow_phys_in)
    real     :: slow_phys_in !! new physical slowness
    call prop_der11Daxis_phys_slow_update2(slow_phys_in)
    prop_der11Daxis_phys_slow_update = 0
  end function

  !! Updates the slowness
  !! <p>
  !! This subroutine or its corresponding function
  !! must be called each time before applying the operator
  subroutine prop_der11Daxis_phys_slow_update2(slow_phys_in)
    real      :: slow_phys_in !! new physical slowness
    slow_phys = slow_phys_in
  end subroutine

  !! Updates the wavelet.
  !! <p>
  !! This function or its corresponding subroutine must 
  !! called once before applying the operator
  integer function prop_der11Daxis_der1_wavelet_update(wavelet_in) 
    real, dimension(nt)  :: wavelet_in !! new wavelet
    call prop_der11Daxis_der1_wavelet_update2(wavelet_in)
    prop_der11Daxis_der1_wavelet_update=0
  end function

  !! Updates the derivative of the wavelet. 
  !! <p>
  !! Computes its derivative and divides by the sampling
  !! <p>
  !! This function or its corresponding function must be 
  !! called once before applying the operator
  subroutine prop_der11Daxis_der1_wavelet_update2(wavelet_in)
    real, dimension(nt)   :: wavelet_in     !! new wavelet
    wavder = wavelet_in
  end subroutine

  !! Applies the forward or adjoint of operator
  integer function prop_der11Daxis_op(adj, add, model, data)
    logical            :: adj   !! forward or adjoint
    logical            :: add   !! add or clear input
    real,dimension(:)  :: model !! perturbation of the slowness
    real,dimension(:)  :: data  !! shifted wavelet scaled by ds
    call prop_der11Daxis_op2(adj, add, model, data)
    prop_der11Daxis_op = 0
  end function

  !! Applies the forward or adjoint of the operator
  subroutine prop_der11Daxis_op2(adj, add, model, data)
    logical                :: adj   !! forward or adjoint
    logical                :: add   !! add or clear input
    real, dimension(:)     :: model !! perturbation of the slowness
    real, dimension(:)     :: data  !! the shifted wavelet scaled by ds
    real                   :: time_prop

    !call adjnull (adj,add,model,data)

    if(adj) then
      time_prop = (dist * slow_phys) - (dly + ot)
      call lint1_stat_init (ow, dt, -time_prop)
      call lint1_stat_lop2 (F, F, wavder, wavder_shift)
      !model(slow_phys_lag)= model(slow_phys_lag) + &
      !  sum(der1_wavelet_shift*data)*(dtdslow/nt)
      model(size(model)) = dot_product(wavder_shift,data) * (dtdslow)
    else
      time_prop = (dist * slow_phys) - (dly + ot)
      call lint1_stat_init (ow, dt, -time_prop)
      call lint1_stat_lop2 (F,F, wavder, wavder_shift)
      !data=data + der1_wavelet_shift*model(slow_phys_lag)*(dtdslow/nt)
      data = wavder_shift * model(size(model)) * (dtdslow)
    end if
  end subroutine

end module
