!! Extended non-linear operator
!! implementation of L~ in notes
!!
!! @author Biondo Biondi
!! @version 2016.07.16
module prop_extend1Daxis_mod

  use sep
  use mem_mod
  use adj_mod
  use axis_mod
  use tcai1_nb  
  use lint1_stat  

  implicit none

  logical, private, parameter              :: T = .true., F = .false.

  ! Axes
  integer                                    :: nt, nw, ntau
  real                                       :: dt, dw, dtau
  real                                       :: ot, ow, otau

  type(axis)                               :: daxis, waxis, caxis
  logical, private                         :: slow_extend          !! check whether to extend slowness
  integer, private                         :: zero_lag             !! position of zero lag
  real, private                            :: dly                  !! time delay of wavelet
  real, private                            :: slow_phys            !! physical background slowness
  real, private                            :: dist                 !! propagation distance
  real, private                            :: weight

  real, dimension(:), allocatable, private :: model_conv          !! extended background slowness (model space)
  real, dimension(:), allocatable, private :: wavelet_shift       !! shifted wavelet
  real, dimension(:), allocatable, private :: wavelet             !! input wavelet
  real, dimension(:), pointer, private     :: wavelet_shift_p     !! shifted wavelet scaled by 1/lag_n

  contains

    !! Initialize the non-linear extended operator
  subroutine prop_extend1Daxis_init(dax_in, wax_in, cax_in, dly_in, dist_in, wav_in)
    type(axis)         :: dax_in, wax_in, cax_in
    real               :: dist_in                !! propagation distance
    real               :: dly_in                 !! time delay of wavelet
    real, dimension(:) :: wav_in

    daxis   = dax_in
    waxis   = wax_in
    caxis   = cax_in
    dly     = dly_in
    dist    = dist_in

    !weight = 1./sqrt(float(caxis%n))
    weight = 1.0

    if (waxis%d .ne. daxis%d) then
      write(0,*)'dtdata=',daxis%d
      write(0,*)'dtwavlet=',waxis%d
      call seperr("dtwavelet must be equal to dtdata")
    end if

    ! Check if slowness is extended
    slow_extend = caxis%n > 1
    if (slow_extend) then
      if (modulo(caxis%o, caxis%d) .ne. 0.) &
        call seperr("tau origin must be a multiple of tau sampling")
      zero_lag = nint(-caxis%o / caxis%d + 1)
      if ((zero_lag < 2) .or. (zero_lag > caxis%n)) &
        call seperr("zero_lag not within lag interval")
    else
      call seperr("Why call prop_extend_1D when slowness is not extended?")
    end if

    ! Size of model space (c)
    call mem_alloc(model_conv, caxis%n - 1)

    ! Read and allocate wavelet
    call mem_alloc(wavelet, waxis%n)
    wavelet = wav_in

    ! Size of shifted wavelet (data)
    call mem_alloc(wavelet_shift, daxis%n)
    allocate(wavelet_shift_p(daxis%n))

  end subroutine

  !! Updates the physical background slowness
  integer function prop_extend1Daxis_back_slow_update(slow_phys_in)
    real                :: slow_phys_in !! new physical background slowness
    call prop_extend1Daxis_back_slow_update2(slow_phys_in)
    prop_extend1Daxis_back_slow_update = 0
  end function

  !! Updates the physical background slowness
  subroutine prop_extend1Daxis_back_slow_update2(slow_phys_in)
    real                :: slow_phys_in !! new physical background slowness
    slow_phys = slow_phys_in
  end subroutine

  !! Updates the wavelet
  integer function prop_extend1Daxis_wavelet_update(wavelet_in)
    real, dimension(:)  :: wavelet_in !! new wavelet
    call prop_extend1Daxis_wavelet_update2(wavelet_in)
    prop_extend1Daxis_wavelet_update = 0
  end function

  !! Updates the wavelet
  subroutine prop_extend1Daxis_wavelet_update2(wavelet_in)
    real, dimension(:)  :: wavelet_in !! new wavelet
    wavelet = wavelet_in
  end subroutine

  !! Applies forward or adjoint of extended non-linear operator
  integer function prop_extend1Daxis_op(adj, add, model, data)
    logical            :: adj   !! apply the forward or the adjoint
    logical            :: add   !! add or clear input
    real,dimension(:)  :: model !! extended slowness
    real,dimension(:)  :: data  !! shifted, filtered wavelet
    call prop_extend1Daxis_op2(adj, add, model, data)
    prop_extend1Daxis_op = 0
  end function

  !! Applies forward or adjoint of extended non-linear operator
  subroutine prop_extend1Daxis_op2(adj, add, model, data)
    logical             :: adj   !! apply the forward or the adjoint
    logical             :: add   !! add or clear the input
    real, dimension(:)  :: model !! extended slowness (c)
    real, dimension(:)  :: data  !! shifted filtered wavelet

    real                :: time_prop

    call adjnull (adj,add,model,data )

    if(adj) then
      time_prop = (dist * slow_phys ) - (dly + daxis%o - caxis%o)
      call lint1_stat_init (waxis%o, daxis%d, -time_prop)
      call lint1_stat_lop2 (F, F, wavelet, wavelet_shift)
      wavelet_shift_p = wavelet_shift * weight
      call tcai1_nb_init(wavelet_shift_p)
      model_conv(1:caxis%n - 1) = model(1:caxis%n - 1)
      call tcai1_nb_lop2(T, add, model_conv, data)
      model(1:caxis%n - 1) = model_conv(1:caxis%n - 1)
      model(zero_lag)      = 0.0
      model(caxis%n)       = model(caxis%n)
    else
      time_prop = (dist * slow_phys) - (dly + daxis%o - caxis%o)
      call lint1_stat_init (waxis%o, daxis%d, -time_prop)
      call lint1_stat_lop2 (F, F, wavelet, wavelet_shift)
      wavelet_shift_p = wavelet_shift * weight
      call tcai1_nb_init(wavelet_shift_p)
      model_conv(1:caxis%n - 1) = model(1:caxis%n - 1)
      model_conv(zero_lag)      = 0.0
      call tcai1_nb_lop2(F, add, model_conv, data)
    end if
  end subroutine

end module
