!! Linearized extended operator
!! <p>
!! implementation of L~^dot in notes
!!
!! @author Biondo Biondi
!! @version 2016.07.15
module prop_extend_der11Daxis_mod

  use axis_mod
  use mem_mod
  use adj_mod
  use tcai1_nb  
  use lint1_stat  
  use igrad1

  implicit none

  logical, private, parameter                :: T = .true., F = .false.

  ! Axes
  integer                                    :: nt, nw, ntau
  real                                       :: dt, dw, dtau
  real                                       :: ot, ow, otau

  integer, private                           :: zlag           !! zero lag location
  integer, private                           :: plag           !! physical lag location
  real, private                              :: dist           !! propagation distance
  real, private                              :: dly            !! initial wavelet delay
  real, private                              :: bslow          !! physical background slowness
  real, private                              :: dtdslow        !! negative propagation distance
  real, private                              :: weight

  real, dimension(:), allocatable, private   :: extslow        !! extended model (model space)
  real, dimension(:), allocatable, private   :: tmpext         !! extended model (model space)
  real, dimension(:), allocatable, private   :: tmpdat
  real, dimension(:), allocatable, private   :: wavder         !! derivative of input wavelet
  real, dimension(:), allocatable, private   :: wavder_shift   !! derivative of shifted wavelet
  real, dimension(:), pointer, private       :: wavder_shift_p !! derivative of scaled shifted wavelet

  contains

    !! Initialized the linearized extended operator
  subroutine prop_extend_der11Daxis_init(daxin, waxin, caxin, dlyin, distin, derin)
    type(axis)         :: daxin, waxin, caxin
    real               :: distin                               !! propagation distance
    real               :: dlyin                          !! time delay of wavelet
    real, dimension(:) :: derin


    nt   = daxin%n; ot   = daxin%o; dt   = daxin%d
    nw   = waxin%n; ow   = waxin%o; dw   = waxin%d
    ntau = caxin%n; otau = caxin%o; dtau = caxin%d

    dly  = dlyin
    dist = distin

    dtdslow = -dist

    call mem_alloc(extslow, ntau - 1)
    call mem_alloc(tmpext , ntau - 1)
    call mem_alloc(tmpdat, nt)

    call mem_alloc(wavder, nw)
    wavder = derin

    !weight = 1./sqrt(float(ntau))
    weight = 1.0

    call mem_alloc(wavder_shift, nt)
    allocate(wavder_shift_p(nt))

    if (dw .ne. dt) then
      write(0,*)'dtdata=',dt
      write(0,*)'dtwavelet=',dw
      call seperr("dtwavelet must be equal to dtdata")
    end if

    write(0,*) otau,dtau,otau/dtau
    if (modulo(otau,dtau) .ne. 0.) then
      call seperr("otau must be a multiple of dtau")
    endif
    zlag=nint(-otau/dtau+1)

    if ((zlag < 2) .or. (zlag > ntau)) then 
      call seperr("zero_lag not within lag interval")
    end if

  end subroutine

  !! Updates the physical background slowness
  integer function prop_extend_der11Daxis_back_slow_update(bslow_in)
    real  :: bslow_in !! new background slowness
    call prop_extend_der11Daxis_back_slow_update2(bslow_in)
    prop_extend_der11Daxis_back_slow_update = 0
  end function

  !! Updates the physical background slowness
  subroutine prop_extend_der11Daxis_back_slow_update2(bslow_in)
    real   :: bslow_in !! new background slowness
    bslow = bslow_in
  end subroutine

  !! Updates the extended slowness
  integer function prop_extend_der11Daxis_extnd_slow_update(extslow_in)
    real, dimension(:) :: extslow_in
    call prop_extend_der11Daxis_extnd_slow_update2(extslow_in)
    prop_extend_der11Daxis_extnd_slow_update = 0
  end function

  !! Updates the extended slowness
  subroutine prop_extend_der11Daxis_extnd_slow_update2(extslow_in)
    real, dimension(:) :: extslow_in
    extslow = extslow_in
    extslow(zlag) = 0.0
  end subroutine

  !! Updates the derivative of the input wavelet
  integer function prop_extend_der11Daxis_der1_wavelet_update(wavelet_in) 
    real, dimension(nw)  :: wavelet_in  !! new wavelet
    call prop_extend_der11Daxis_der1_wavelet_update2(wavelet_in)
    prop_extend_der11Daxis_der1_wavelet_update=0
  end function

  !! Updates the derivative of the input wavelet
  !! <p>
  !! Computes the derivative of the wavelet and divides by the sampling
  subroutine prop_extend_der11Daxis_der1_wavelet_update2(wavelet_in)
    real, dimension(nw)   :: wavelet_in
    wavder = wavelet_in
  end subroutine

  !! Applies the linearized extended operator
  integer function prop_extend_der11Daxis_op(adj, add, model, data)
    logical            :: adj   !! apply adjoint or forward
    logical            :: add   !! add or clear input
    real,dimension(:)  :: model !! extended slowness (c)
    real,dimension(:)  :: data  !! derivative of wavelet convolved with c
    call prop_extend_der11Daxis_op2(adj, add, model, data)
    prop_extend_der11Daxis_op = 0
  end function

  !! Applies the linearized extended operator
  subroutine prop_extend_der11Daxis_op2(adj, add, model, data)
    logical                   :: adj
    logical                   :: add
    real, dimension(:)        :: model
    real, dimension(:)        :: data

    real                      :: time_prop

    call adjnull (adj,add,model,data )

    if(adj) then
      tmpext = 0
      time_prop = (dist * bslow) - (dly + ot - otau)
      call lint1_stat_init (ow, dt, -time_prop)
      call lint1_stat_lop2 (F, F, wavder, wavder_shift)
      wavder_shift_p = wavder_shift * (dtdslow * weight)
      call tcai1_nb_init(wavder_shift_p)
      call tcai1_nb_lop2(T, F, tmpext, data)
      model(size(model)) = model(size(model)) + dot_product(extslow, tmpext)
    else
      tmpdat = 0
      time_prop = (dist * bslow) - (dly + ot - otau)
      call lint1_stat_init (ow, dt, -time_prop)
      Call lint1_stat_lop2 (F, F, wavder, wavder_shift)
      wavder_shift_p = wavder_shift * (dtdslow * weight)
      call tcai1_nb_init(wavder_shift_p)
      call tcai1_nb_lop2(F, F, extslow, tmpdat)
      data = data + tmpdat * model(size(model))
    end if
  end subroutine

end module
