module solver_nlcg_mod
! Nonlinear CG solver
! Last edited by Ali Almomin 08/24/2015

  use solver_obj_mod
  use prob_obj_mod
  use stepper_obj_mod
  use terminate_obj_mod

  implicit none

  private
  public :: solver_nlcg

  type, extends(solver_obj)         :: solver_nlcg !extending solver to NLCG
    !Variable extension
    character(20)                   :: beta_type='FR'
    contains
    procedure                       :: run      !Function to run the solver
  end type

  contains

  integer function run(this, modl, prob, stpr, term)
    class(solver_nlcg)              :: this
    !-------------------- interface variables -------
    integer                         :: stat
    real, dimension(:)              :: modl            ! model (input/output)
    class(prob_obj)                 :: prob
    class(stepper_obj)              :: stpr
    class(terminate_obj)            :: term
    double precision                :: beta, alpha, f
    integer                         :: iter
    logical                         :: success
    real, dimension(:), allocatable :: grad
    real, dimension(:), allocatable :: grad0
    real, dimension(:), allocatable :: dmodl

    procedure(betaFR), pointer      :: betaf

    select case (this%beta_type)
      case ('FR')
        betaf => betaFR
      case ('PRP')
        betaf => betaPRP
      case ('HS')
        betaf => betaHS
      case ('CD')
        betaf => betaCD
      case ('LS')
        betaf => betaLS
      case ('DY')
        betaf => betaDY
      case ('N')
        betaf => betaN
      case default
        write(0,*)'Not a beta type.'
       call exit(0)
    end select

    allocate(grad(prob%nmodl))
    allocate(grad0(prob%nmodl))
    allocate(dmodl(prob%nmodl))

    dmodl = 0.
    beta = 0.
    iter = 1
    do
      grad0 = grad
      stat = prob%get_grad(modl, grad)
      stat = prob%output(modl)
      if(iter > 1) beta = betaf(grad, grad0, dmodl)
      if(beta < 0.) beta = 0.
      dmodl = beta*dmodl - grad ! search direction
      stat = stpr%run(modl, prob, dmodl, alpha, success)
      if(.not. success) then
        write(0,*) "Stepper couldn't find a proper step size, will terminate solver"
        exit
      end if
      stat = prob%get_obj(modl, f)
      write(0,*) "iter", iter, " obj =", real(f), " modl=", modl(prob%nmodl), &
                 " grad =", grad(prob%nmodl), " alpha =", alpha, " feval =", prob%get_fevals()
      iter = iter + 1
      if(term%test(iter, prob)) exit
    end do

    deallocate(grad)
    deallocate(grad0)
    deallocate(dmodl)

    run = 0
  end function

! beta functions
! g= new gradient, g0=old, d= direction
! From A SURVEY OF NONLINEAR CONJUGATE GRADIENT METHODS

  !Fletcher and Reeves
  double precision function betaFR(g,g0,d)
    real, dimension(:)  :: g, g0,d
    betaFR = sum(dprod(g,g))/sum(dprod(g0,g0))
  end function

  ! Polak, Ribiere, Polyak
  double precision function betaPRP(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaPRP = sum(dprod(g,g-g0))/sum(dprod(g0,g0))
  end function

  ! Hestenes and Stiefel
  double precision function betaHS(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaHS = sum(dprod(g,g-g0))/sum(dprod(d,g-g0))
  end function

  ! Conjugate Descent
  double precision function betaCD(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaCD = -sum(dprod(g,g))/sum(dprod(d,g0))
  end function

  ! Liu and Storey
  double precision function betaLS(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaLS = -sum(dprod(g,g-g0))/sum(dprod(d,g0))
  end function

  ! Dai and Yuan
  double precision function betaDY(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaDY = sum(dprod(g,g))/sum(dprod(d,g-g0))
  end function

  ! Hager and Zhang
  double precision function betaN(g,g0,d)
    real, dimension(:)  :: g, g0, d
    betaN = sum(dprod(g-g0-2*sum(dprod(g-g0,g-g0))*d/sum(dprod(d,g-g0)),g))/sum(dprod(d,g-g0))
  end function

end module
